/* M E T A X P O N   a u d i o	 t i m e - s c a l i n g   l i b r a r y
 *
 * Portable edition
 * Supports single precision floating point data
 */
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <limits.h>
#include <math.h>

/* Constants, macros, and FFT helper functions */

#define FFTBITS			12
#define FFTSIZE			(1 << FFTBITS)
#define PI			3.1415927f

#if !defined DJBFFT && !defined SPL && !defined IPP

typedef struct {
	float re, im;
} Complex;

#endif

#if defined FFTW || defined SFFT || defined VDSP

static void scale(Complex *a)
{
	int i;

	for (i = 0; i < FFTSIZE; i++) {
		a[i].re /= FFTSIZE;
		a[i].im /= FFTSIZE;
	}
}

#endif

#ifdef DJBFFT		/* D. Bernstein's djbfft */

#include <fftc4.h>
#include <fftfreq.h>
#define Complex		complex4
#define fft(x, y)	(fftc4_4096(x), fftc4_scale4096(x))
#define ffti(x, y)	fftc4_un4096(x)

static unsigned freq[FFTSIZE];

#elif defined FFTW	/* "Fastest Fourier Transform in the West" (M. Frigo & S. Johnson) */

#include <fftw3.h>
#define fft(x, y)	 fftwf_execute(PlanF)
#define ffti(x, y)	(fftwf_execute(PlanB), scale(x))

fftwf_plan PlanF, PlanB;

static void freePlans(void)
{
	fftwf_destroy_plan(PlanF);
	fftwf_destroy_plan(PlanB);
	fftwf_cleanup();
}

#elif defined SFFT	/* A. Blake's SFFT */

#include <sfft.h>
#define fft(x, y)	(sfft_execute(PlanF, x, FFTbuf), memcpy(x, FFTbuf, FFTSIZE * sizeof(Complex)))
#define ffti(x, y)	(sfft_execute(PlanB, x, FFTbuf), memcpy(x, FFTbuf, FFTSIZE * sizeof(Complex)), scale(x))

sfft_plan_t *PlanF, *PlanB;

/* SFFT doesn't work in-place; we need a properly aligned buffer (other compilers?!) */

#  if defined __SUNPRO_C && __SUNPRO_C >= 0x400
#    pragma align 32 (FFTbuf)
#  elif defined _MSC_VER && _MSC_VER >= 1310 || defined __INTEL_COMPILER && __INTEL_COMPILER >= 400
     __declspec(align(32))
#  elif defined __GNUC__ && __GNUC__ >= 3
     __attribute__ ((aligned(32)))
#  endif

static Complex FFTbuf[FFTSIZE];

static void freePlans(void)
{
	sfft_free(PlanF);
	sfft_free(PlanB);
}

#elif defined MLIB	/* Sun mediaLib */

#include <mlib.h>
#define fft(x, y)	mlib_SignalFFT_1_F32C((mlib_f32 *)(x), FFTBITS)
#define ffti(x, y)	mlib_SignalIFFT_1_F32C((mlib_f32 *)(x), FFTBITS)

#elif defined SPL	/* Sun Performance Library */

#include <sunperf.h>
#define Complex		complex
#define re		r
#define im		i
#define fft(x, y)	(iopt = (y), scale = 1.f, cfftc_(&iopt, &len, &scale, (x), (x), trigs, ifac, work, &lwork, &ierr))
#define ffti(x, y)	(iopt = (y), scale = 1.f / FFTSIZE, cfftc_(&iopt, &len, &scale, (x), (x), trigs, ifac, work, &lwork, &ierr))
#pragma align 32 (ifac, trigs, work)

static int iopt, len = FFTSIZE, ifac[128], lwork = 2 * FFTSIZE, ierr;
static float scale, trigs[2 * FFTSIZE], work[2 * FFTSIZE];

#elif defined VDSP	/* Apple vDSP (Accelerate/vecLib framework) */

#include <Accelerate/Accelerate.h>
#define fft(x, y)	(move(x, 0), vDSP_fft_zipt(fftSetup, &fftData, 1, &fftTemp, FFTBITS, kFFTDirection_Forward), move(x, 1))
#define ffti(x, y)	(move(x, 0), vDSP_fft_zipt(fftSetup, &fftData, 1, &fftTemp, FFTBITS, kFFTDirection_Inverse), move(x, 1), scale(x))
#define ALIGN32		__attribute__ ((aligned(32)))

static FFTSetup fftSetup;
static DSPSplitComplex fftData, fftTemp;
static float dataReal[FFTSIZE] ALIGN32, dataImag[FFTSIZE] ALIGN32, tempReal[FFTSIZE] ALIGN32, tempImag[FFTSIZE] ALIGN32;

static void move(Complex *p, int dir)
{
	int i;

	if (dir)
		for (i = 0; i < FFTSIZE; i++) {
			p[i].re = dataReal[i];
			p[i].im = dataImag[i];
		}
	else for (i = 0; i < FFTSIZE; i++) {
			dataReal[i] = p[i].re;
			dataImag[i] = p[i].im;
		}
}

static void freevDSP(void)
{
	vDSP_destroy_fftsetup(fftSetup);
}

#elif defined ACML	/* AMD Core Math Library */

#include <acml.h>
#define Complex		complex
#define re		real
#define im		imag
#define fft(x, y)	cfft1d(y, FFTSIZE, x, comm, &info);
#define ffti(x, y)	cfft1d(y, FFTSIZE, x, comm, &info);

static complex comm[5 * FFTSIZE + 100];
static int info;

#elif defined IPP	/* Intel Integrated Performance Primitives */

#include <ipp.h>
#define Complex		Ipp32fc
#define fft(x, y)	ippsFFTFwd_CToC_32fc_I((x), fftSpec, fftBuf)
#define ffti(x, y)	ippsFFTInv_CToC_32fc_I((x), fftSpec, fftBuf)

static IppsFFTSpec_C_32fc *fftSpec;
static Ipp8u *fftBuf;

static void freeIPP(void)
{
	ippsFFTFree_C_32fc(fftSpec);
	free(fftBuf);
}

#else

#define ffti fft

/* Built-in FFT: slow */

static void fft(Complex *c, int sign)
{
	int CurBlockSize, i, j, k, l;
	Complex q, w, e, r, t;

	/* Bit reversal, first and last unchanged */
	for (j = FFTSIZE / 2, i = 1; i < FFTSIZE - 1; i++, j += k) {
		if (i < j) {
			t = c[i];
			c[i] = c[j];
			c[j] = t;
		}
		for (k = FFTSIZE / 2; k <= j; k /= 2)
			j -= k;
	}
	/* Composition */
	for (i = 0; i < FFTSIZE; i += 4) {
		r.re = sign > 0 ? c[i + 3].im - c[i + 2].im : c[i + 2].im - c[i + 3].im;
		r.im = sign > 0 ? c[i + 2].re - c[i + 3].re : c[i + 3].re - c[i + 2].re;
		q.re = c[i].re + c[i + 1].re;
		q.im = c[i].im + c[i + 1].im;
		w.re = c[i].re - c[i + 1].re;
		w.im = c[i].im - c[i + 1].im;
		e.re = c[i + 2].re + c[i + 3].re;
		e.im = c[i + 2].im + c[i + 3].im;
		c[i].re = q.re + e.re;
		c[i].im = q.im + e.im;
		c[i + 2].re = q.re - e.re;
		c[i + 2].im = q.im - e.im;
		c[i + 1].re = w.re + r.re;
		c[i + 1].im = w.im + r.im;
		c[i + 3].re = w.re - r.re;
		c[i + 3].im = w.im - r.im;
	}
	for (CurBlockSize = 4, k = 0; k < FFTBITS - 2; k++, CurBlockSize *= 2) {
		float shift_r, shift_i, step_r, step_i, arg = PI / CurBlockSize;
		int CurBlockSize2 = CurBlockSize * 2;

		for (i = 0; i < FFTSIZE; i += CurBlockSize2) {
			t = c[i + CurBlockSize];
			c[i + CurBlockSize].re = c[i].re - t.re;
			c[i + CurBlockSize].im = c[i].im - t.im;
			c[i].re += t.re;
			c[i].im += t.im;
		}
		shift_r = step_r = (float)cos(arg);
		shift_i = step_i = (float)(sign * sin(arg));
		for (j = 1; ; ) {
			for (i = j; i < FFTSIZE; i += CurBlockSize2) {
				l = i + CurBlockSize;
				t.re = c[l].re * shift_r - c[l].im * shift_i;
				t.im = c[l].re * shift_i + c[l].im * shift_r;
				c[l].re = c[i].re - t.re;
				c[l].im = c[i].im - t.im;
				c[i].re += t.re;
				c[i].im += t.im;
			}
			if (++j >= CurBlockSize)
				break;
			t.re = shift_r * step_r - shift_i * step_i;
			shift_i = shift_r * step_i + shift_i * step_r;
			shift_r = t.re;
		}
	}
	/* Normalisation */
	if (sign < 0)
		for (i = 0; i < FFTSIZE; i++) {
			c[i].re /= FFTSIZE;
			c[i].im /= FFTSIZE;
		}
}

#endif

#define IN_LIB_1
#include "metaxpon.h"

/* Phase vocoder definitions */

#ifndef PRO
#  define PRO			0	/* 0: FAST method, 1: PRO method */
#endif
#define OVERLAP			2
#define OVERSAMPBITS		(OVERLAP + PRO)
#define OVERSAMP		(1 << OVERSAMPBITS)
#define HOPSIZE			(FFTSIZE / OVERSAMP)
#define HOPSIZE_SHIFT		(FFTBITS - OVERSAMPBITS)	/* log2(HOPSIZE) */

#define ATTACK_LOCK		4	/* Blocks after attack that can't have attack */
#define ATTACK_BLOCK		1	/* Number of successive blocks with input phase */
#define ATTACK_AMPLIFY		1	/* 1: use amplification at attacks, 0: don't use */
#define ATTACK_AMP_LEVEL	1.4f	/* Attack amplification level */
#define ATTACK_TEMPO_LOCK	0	/* Number of successive blocks with tempo locked to 100% */

#define ATTACK_PHASE_BLOCK(x)	(x > ATTACK_LOCK - ATTACK_BLOCK)
#define ATTACK_TEMPO_BLOCK(x)	(x > ATTACK_LOCK - ATTACK_TEMPO_LOCK)

#ifndef max
#  define max(a,b) ((a) > (b) ? (a) : (b))
#endif

/* Phase modification */

static void Vocode(float *MagnS, float *Real, float *Imag, float *OldInPhase, float *OutPhase, int *Peak, unsigned InStep, float *Energy, short *Attack, float Level) /* energy ratio */
{
	int i, j, OldPeak = 0, MinI = 0, CurPeak = -1;
	float MinV, tr, zr = 1, zi = 0;

	/* Peak detection */
	Energy[2] = Energy[1];
	Energy[1] = Energy[0];
	Energy[0] = 0;
	for (i = 0; i <= FFTSIZE / 2; i++)
		Energy[0] += MagnS[i] = Real[i] * Real[i] + Imag[i] * Imag[i];
	if (!*Attack && Energy[0] > Energy[1] && Energy[1] > Energy[2] && Energy[0] > Energy[2] * Level)
		*Attack = ATTACK_LOCK;
	else if (*Attack)
		(*Attack)--;
	OldInPhase[0] = OutPhase[0] = Real[0] > 0 ? 0 : PI;
	for (MinV = MagnS[0], i = 2; i < FFTSIZE / 2 - 1; i++) {
		if (MagnS[i] < MinV) {
			MinV = MagnS[i];
			MinI = i;
		}
		if (MagnS[i] > max(max(MagnS[i - 1], MagnS[i + 1]), max(MagnS[i - 2], MagnS[i + 2]))) {
			/* Peak phase */
			float InPhase = (float)atan2(Imag[i], Real[i]);
			if (ATTACK_PHASE_BLOCK(*Attack))
				OutPhase[i] = InPhase;
			else {
				float Delta = (float)fmod(InPhase - OldInPhase[Peak[i]] - InStep * i * 2 * PI / FFTSIZE, 2 * PI);
				if (Delta < -PI)
					Delta += 2 * PI;
				else if (Delta > PI)
					Delta -= 2 * PI;
				OutPhase[i] = (float)fmod(OutPhase[Peak[i]] + HOPSIZE * (i * 2 * PI / FFTSIZE + Delta / InStep), 2 * PI);
			}
			OldInPhase[i] = InPhase;
			/* Mark peak area */
			for (j = OldPeak; j <= MinI; j++)
				Peak[j] = OldPeak;
			OldPeak = i;
			for (j = MinI + 1; j < i; j++)
				Peak[j] = OldPeak;
			MinV = MagnS[i];
			MinI = i;
		}
	}
	for (j = OldPeak; j < FFTSIZE / 2; j++)
		Peak[j] = OldPeak;
	for (i = 1; i < FFTSIZE / 2; i++) { /* 0 and max. freq. don't contain phase information */
		if (CurPeak != Peak[i]) {
			CurPeak = Peak[i];
			zr = (float)cos(OutPhase[CurPeak] - OldInPhase[CurPeak]);
			zi = (float)sin(OutPhase[CurPeak] - OldInPhase[CurPeak]);
		}
		tr = Real[i] * zr - Imag[i] * zi;
		Imag[i] = Real[i] * zi + Imag[i] * zr;
		Real[i] = tr;
	}
}

/* Data structure-interface between application programmes and the library */

#if FLOAT_DATA
#  define IODATATYPE	float
#else
#  define IODATATYPE	short
#endif

#pragma pack(1)
struct VocodeBuffer {	/* Corresponds to the public structure MXdata */
	DWORD Version;	/* Fixed point 16.16 */
	DWORD Padding;	/* Align next pointers to 8 bytes to avoid crash on Solaris 64-bit SPARC */
	IODATATYPE *InBuff; /* [INBLOCKSAM*2]	Input buffer address */
	IODATATYPE *OutBuff;/* [OUTBLOCKSAM*n] Output buffer address */

	Complex *Comp;	    /* [FFTSIZE]	Calc buffers (MXdata.Memory) */
	float *Window;	    /* [FFTSIZE]	Window function */
	float *OutL;	    /* [FFTSIZE]	Mixing output */
	float *OutR;	    /* [FFTSIZE]	"   "	"   " */
	float *OldInPhaseL; /* [FFTSIZE/2]	Signal phases */
	float *OutPhaseL;   /* [FFTSIZE/2]	"   "	"   " */
	float *OldInPhaseR; /* [FFTSIZE/2]	"   "	"   " */
	float *OutPhaseR;   /* [FFTSIZE/2]	"   "	"   " */
	int   *PeakL;	    /* [FFTSIZE/2]	Spectrum peak values */
	int   *PeakR;	    /* [FFTSIZE/2]	"   "	"   "	"    */
	float *MagnS;	    /* [FFTSIZE/2+4]	Calculation buffer */
	float *RealA;	    /* [FFTSIZE/2+4]	"   "	"   "	"  */
	float *ImagA;	    /* [FFTSIZE/2+4]	"   "	"   "	"  */
	float *RealB;	    /* [FFTSIZE/2+4]	"   "	"   "	"  */
	float *ImagB;	    /* [FFTSIZE/2+4]	"   "	"   "	"  */
	float *EnergyL;	    /* [3]		Energy of blocks */
	float *EnergyR;	    /* [3]		"   "	"   "	 */

	DWORD InBuffOfs;    /*			Current position in input buffer */
	DWORD OutBuffOfs;   /*			"   "	"   "	"  output buffer */
	short AttackL;	    /*			Blocks with attack */
	short AttackR;	    /*			"   "	"   "	"  */

	DWORD InBlkSz;	    /*	Input buffer size [bytes] */
	DWORD OutBlkSz;	    /* Output buffer size [bytes] */
	DWORD QuadInBlSamsk;/* (4*INBLOCKSAM-1)*IODATATYPE (mask) */
	DWORD QuadInBlSmask;/*	4*INBLOCKSAM-1 (mask) */
	DWORD DualInBlSmask;/*	2*INBLOCKSAM-1 (mask) */
	DWORD DualInblSam;  /*	2*INBLOCKSAM */
	BYTE  InBlkSamShInc;/*	  INBLOCKSAM_shift+1 */
	BYTE  OutBlkHops;   /*	 OUTBLOCKSAM/HOPSIZE */
		short dummy;/* align 4 */

	DWORD Speed;	    /* Tempo of playback, fixed point 16.16 (MXdata.Tempo) */
	DWORD Threshold;    /* Attack detection level, fixed point 16.16 */
	DWORD Last;	    /* Bytes to read last before data end (multiply of sample size) */
	BYTE  Empty;	    /* Set when ready to fill input buffer */
	BYTE  Half;	    /* Currently played half of the input buffer (0|1) */
	BYTE  E0F;	    /* 0: no data end, 1: end in the 1st half, 2: end in the 2nd half */
	BYTE  LastFiledBuf; /* Last filled output buffer half (0|1) MXdata.OutBufNum */
	BYTE  OutBufferSzKB;/* Output buffer size [KB]: 2*|4*|8|16|32 (*sometimes) MXdata.OutBufSizeK */
	BYTE  fpu;	    /* 0: use FPU, 1: use 3DNow!, 2: use SSE (x86 only) MXdata.MathUnit */
};
#pragma pack()

#undef	IN_LIB_1
#define IN_LIB_2
#define MXdata VocodeBuffer
#include "metaxpon.h"

/*** Public functions ***/

/* Main vocoder function. On end of the input stream, return the count of the
 * stereo samples written, else return -1.
 * vb->Speed = 0x10000: normall speed, 0x8000: 2x slower, 0x20000: 2x faster.
 */
int __cdecl MXprocess(struct VocodeBuffer *vb)
{
	unsigned i, j, k, InStep = (unsigned int)((HOPSIZE * vb->Speed + 0x8000) / 0x10000);
	DWORD OldInBuffOfs = vb->InBuffOfs;

	if (vb->E0F && !(vb->E0F & 0x40)) {
		memset(vb->InBuff + ((vb->E0F - 1) * vb->InBlkSz + vb->Last) / sizeof(IODATATYPE), 0, vb->InBlkSz - vb->Last);
		vb->E0F |= 0x40;
	}
	for (j = 0; j < (unsigned)vb->OutBlkHops; j++) { /* For each hop till output buffer is full */
#ifdef DJBFFT
		int l, m;
#endif
		/* Read */
		IODATATYPE *ib = vb->InBuff + vb->InBuffOfs;
		for (i = 0; i < (FFTSIZE >> PRO); i++) {
			vb->Comp[i].re = *ib++ * vb->Window[i];
			vb->Comp[i].im = *ib++ * vb->Window[i];
			if (ib > vb->InBuff + vb->QuadInBlSmask)
				ib -= vb->QuadInBlSmask + 1; /* 4 * INBLOCKSAM */
#if PRO
			vb->Comp[i + FFTSIZE / 2].re = vb->Comp[i + FFTSIZE / 2].im = 0;
#endif
		}
		/* Process */
		fft(vb->Comp, -1);
		vb->RealA[0] = vb->Comp[0].re * 2;
		vb->RealB[0] = vb->Comp[0].im * 2;
		vb->ImagA[0] = vb->ImagB[0] = 0;
		for (k = FFTSIZE - 1, i = 1; i <= FFTSIZE / 2; i++, k--) {
#ifdef DJBFFT
			l = freq[i];
			m = freq[k];
			vb->RealA[i] = vb->Comp[l].im + vb->Comp[m].im;
			vb->RealB[i] = vb->Comp[l].re + vb->Comp[m].re;
			vb->ImagA[i] = vb->Comp[l].re - vb->Comp[m].re;
			vb->ImagB[i] = vb->Comp[m].im - vb->Comp[l].im;
#else
			vb->RealA[i] = vb->Comp[i].re + vb->Comp[k].re;
			vb->RealB[i] = vb->Comp[i].im + vb->Comp[k].im;
			vb->ImagA[i] = vb->Comp[i].im - vb->Comp[k].im;
			vb->ImagB[i] = vb->Comp[k].re - vb->Comp[i].re;
#endif
		}
		Vocode(vb->MagnS, vb->RealA, vb->ImagA, vb->OldInPhaseL, vb->OutPhaseL, vb->PeakL, InStep, vb->EnergyL, &vb->AttackL, (float)vb->Threshold / 0x10000);
		Vocode(vb->MagnS, vb->RealB, vb->ImagB, vb->OldInPhaseR, vb->OutPhaseR, vb->PeakR, InStep, vb->EnergyR, &vb->AttackR, (float)vb->Threshold / 0x10000);
		vb->Comp[0].re = vb->RealA[0] * 2;
		vb->Comp[0].im = vb->RealB[0] * 2;
#ifdef DJBFFT
		for (k = FFTSIZE - 1, i = 1; i < FFTSIZE / 2; i++, k--) {
			l = freq[i];
			vb->Comp[l].im = vb->RealA[i] - vb->ImagB[i];
			vb->Comp[l].re = vb->RealB[i] + vb->ImagA[i];
			l = freq[k];
			vb->Comp[l].im = vb->RealA[i] + vb->ImagB[i];
			vb->Comp[l].re = vb->RealB[i] - vb->ImagA[i];
		}
		l = freq[i];
		vb->Comp[l].im = vb->RealA[FFTSIZE / 2] * 2;
		vb->Comp[l].re = vb->RealB[FFTSIZE / 2] * 2;
#else
		for (k = FFTSIZE - 1, i = 1; i < FFTSIZE / 2; i++, k--) {
			vb->Comp[i].re = vb->RealA[i] - vb->ImagB[i];
			vb->Comp[i].im = vb->RealB[i] + vb->ImagA[i];
			vb->Comp[k].re = vb->RealA[i] + vb->ImagB[i];
			vb->Comp[k].im = vb->RealB[i] - vb->ImagA[i];
		}
		vb->Comp[FFTSIZE / 2].re = vb->RealA[FFTSIZE / 2] * 2;
		vb->Comp[FFTSIZE / 2].im = vb->RealB[FFTSIZE / 2] * 2;
#endif
		ffti(vb->Comp, 1);
		/* Write */
#if ATTACK_AMPLIFY
		if (ATTACK_PHASE_BLOCK(vb->AttackL) || ATTACK_PHASE_BLOCK(vb->AttackR)) {
			float LevelL = 1.f, LevelR = 1.f;

			if (ATTACK_PHASE_BLOCK(vb->AttackL))
				LevelL = ATTACK_AMP_LEVEL;
			if (ATTACK_PHASE_BLOCK(vb->AttackR))
				LevelR = ATTACK_AMP_LEVEL;
			for (i = 0; i < (FFTSIZE >> PRO); i++) {
				vb->OutL[(vb->OutBuffOfs + i) % FFTSIZE] += vb->Comp[i].re * vb->Window[i] * LevelL;
				vb->OutR[(vb->OutBuffOfs + i) % FFTSIZE] += vb->Comp[i].im * vb->Window[i] * LevelR;
			}
		} else
#endif
			for (i = 0; i < (FFTSIZE >> PRO); i++) {
				vb->OutL[(vb->OutBuffOfs + i) % FFTSIZE] += vb->Comp[i].re * vb->Window[i];
				vb->OutR[(vb->OutBuffOfs + i) % FFTSIZE] += vb->Comp[i].im * vb->Window[i];
			}
		/* Write hop to output */
		for (i = 0; i < HOPSIZE; i++) {
#if FLOAT_DATA
			vb->OutBuff[vb->LastFiledBuf * vb->OutBlkSz / sizeof(IODATATYPE) + 2 * HOPSIZE * j + 2 * i] = vb->OutL[vb->OutBuffOfs + i];
			vb->OutBuff[vb->LastFiledBuf * vb->OutBlkSz / sizeof(IODATATYPE) + 2 * HOPSIZE * j + 2 * i + 1] = vb->OutR[vb->OutBuffOfs + i];
#else
			short r;

			if (vb->OutL[vb->OutBuffOfs + i] >= SHRT_MAX)
				r = SHRT_MAX;
			else if (vb->OutL[vb->OutBuffOfs + i] <= SHRT_MIN)
				r = SHRT_MIN;
			else
				r = (short)vb->OutL[vb->OutBuffOfs + i];
			vb->OutBuff[vb->LastFiledBuf * vb->OutBlkSz / sizeof(IODATATYPE) + 2 * HOPSIZE * j + 2 * i] = r;
			if (vb->OutR[vb->OutBuffOfs + i] >= SHRT_MAX)
				r = SHRT_MAX;
			else if (vb->OutR[vb->OutBuffOfs + i] <= SHRT_MIN)
				r = SHRT_MIN;
			else
				r = (short)vb->OutR[vb->OutBuffOfs + i];
			vb->OutBuff[vb->LastFiledBuf * vb->OutBlkSz / sizeof(IODATATYPE) + 2 * HOPSIZE * j + 2 * i + 1] = r;
#endif
			vb->OutL[vb->OutBuffOfs + i] = vb->OutR[vb->OutBuffOfs + i] = 0;
		}
		/* Update pointers */
		vb->OutBuffOfs += HOPSIZE;
		vb->OutBuffOfs &= FFTSIZE - 1;
#if ATTACK_TEMPO_LOCK
		vb->InBuffOfs += 2 * (ATTACK_TEMPO_BLOCK(vb->AttackL) || ATTACK_TEMPO_BLOCK(vb->AttackR) ? HOPSIZE : InStep);
#else
		vb->InBuffOfs += 2 * InStep;
#endif
		vb->InBuffOfs &= vb->QuadInBlSmask;
		vb->Half = (BYTE)(vb->InBuffOfs >> vb->InBlkSamShInc);
		if ((vb->Half + 1) & vb->E0F) {
			vb->E0F |= 0x80;
			if ((vb->InBuffOfs & vb->DualInBlSmask) * sizeof(IODATATYPE) >= vb->Last)
				goto ret;
		} else if (vb->E0F & 0x80) {
ret:			return HOPSIZE * (j + 1);	/* End. */
		}
	}
	/* Input flags */
	if ((vb->InBuffOfs ^ OldInBuffOfs) & vb->DualInblSam) {
		if (vb->E0F) {
			if (vb->Half + 1 == vb->E0F)
				memset(vb->InBuff + (OldInBuffOfs & vb->DualInblSam), 0, vb->InBlkSz);
		} else
			vb->Empty = 1;
	}
	return -1;
}

/* Vocoder variable and buffer initialisation. Call before each start of playback
 * or processing. The OutBufferSzKB field must already contain a valid value.
 */
int __cdecl MXstart(struct VocodeBuffer *vb)
{
	BYTE shift, kb = vb->OutBufferSzKB;
	DWORD bytes = kb * 1024;

	if ((kb != 2 && kb != 4 && kb != 8 && kb != 16 && kb != 32) || !(kb >> (HOPSIZE_SHIFT + 2 + FLOAT_DATA - 10)))
		return -1;	/* buffer size not 2|4|8|16|32 KB or too small for the given HOPSIZE */
	vb->OutBlkSz = bytes;	/* Output buffer size [bytes] */
	vb->InBlkSz = bytes * 4;/*  Input buffer size [bytes] */
	vb->OutBlkHops = (BYTE)(bytes >> (2 + FLOAT_DATA + HOPSIZE_SHIFT)); /* OUTBLOCKSAM / HOPSIZE */
	vb->DualInblSam = bytes <<= (1 - FLOAT_DATA);  /* 2 * INBLOCKSAM */
	for (shift = 0; !((bytes >> shift) & 1); shift++)
		;
	vb->InBlkSamShInc = shift;		       /*     INBLOCKSAM_shift + 1  */
	vb->DualInBlSmask = bytes - 1;		       /* 2 * INBLOCKSAM - 1 (mask) */
	vb->QuadInBlSmask = bytes = 2 * bytes - 1;     /* 4 * INBLOCKSAM - 1 (mask) */
	vb->QuadInBlSamsk = bytes << (1 + FLOAT_DATA); /*(4 * INBLOCKSAM - 1) * IODATATYPE (mask) */

	/* Clear buffers, offsets, attack locks and input buffer half */
	memset(vb->OutL, 0, FFTSIZE * sizeof(float));
	memset(vb->OutR, 0, FFTSIZE * sizeof(float));
	memset(vb->OldInPhaseL, 0, FFTSIZE * sizeof(float) / 2);
	memset(vb->OldInPhaseR, 0, FFTSIZE * sizeof(float) / 2);
	memset(vb->OutPhaseL, 0, FFTSIZE * sizeof(float) / 2);
	memset(vb->OutPhaseR, 0, FFTSIZE * sizeof(float) / 2);
	memset(vb->PeakL, 0, FFTSIZE * sizeof(int) / 2);
	memset(vb->PeakR, 0, FFTSIZE * sizeof(int) / 2);
	vb->InBuffOfs = vb->OutBuffOfs = vb->AttackL = vb->AttackR = vb->Half = 0;
	return 0;
}

static void *align32(void *p)	/* Address alignment to 32 bytes */
{
	union {
		void *q;
		ptrdiff_t n;
	} address;

	address.q = p;
	address.n = (address.n + 31) & -32;
	return address.q;
}

/* Constant data initialisation. Call only once, before any other function here.
 * The InBuff, OutBuff & Comp fields must already contain valid memory addresses.
 */
int __cdecl MXinit(struct VocodeBuffer *vb)
{
	float *p;	/* the first array address - a base for the next ones */
	int i;

	if (vb->Version != METAXPON_VERSION)
		return -1;
	vb->Threshold = 0x20000; /* Default: 2.0 */

	/* Align buffer addresses to 32 bytes */
	vb->InBuff	= align32(vb->InBuff);
	vb->OutBuff	= align32(vb->OutBuff);
	vb->Comp	= align32(vb->Comp);
	p = (float *)vb->Comp;

	/* Set up array addresses (each increment = size of previous buffer) */
	vb->Window	= p +=	FFTSIZE * 2;
	vb->OutL	= p +=	FFTSIZE;
	vb->OutR	= p +=	FFTSIZE;
	vb->OldInPhaseL = p +=	FFTSIZE;
	vb->OutPhaseL	= p +=	FFTSIZE / 2;
	vb->OldInPhaseR = p +=	FFTSIZE / 2;
	vb->OutPhaseR	= p +=	FFTSIZE / 2;
	vb->PeakL	= (int *)(p += FFTSIZE / 2);
	vb->PeakR	= (int *)(p += FFTSIZE / 2 * sizeof(int) / sizeof(float));
	vb->MagnS	= p +=	FFTSIZE / 2 * sizeof(int) / sizeof(float);
	vb->RealA	= p += (FFTSIZE / 2 + 4);
	vb->ImagA	= p += (FFTSIZE / 2 + 4);
	vb->RealB	= p += (FFTSIZE / 2 + 4);
	vb->ImagB	= p += (FFTSIZE / 2 + 4);
	vb->EnergyL	= p += (FFTSIZE / 2 + 4);
	vb->EnergyR	= p + 3;

	/* Initialise the [Blackman] window function data */
	for (i = 0; i < (FFTSIZE >> PRO); i++) {
		float n = .08f * (float)cos(4 * PI * (i + .5) / (FFTSIZE >> PRO));
		vb->Window[i] = 0.63f * (.42f - .5f * (float)cos(2 * PI * (i + .5) / (FFTSIZE >> PRO)) + n);
	}
#ifdef DJBFFT
	for (i = 0; i < FFTSIZE; i++)
		freq[fftfreq_c(i, FFTSIZE)] = i;
#elif defined FFTW
	PlanF = fftwf_plan_dft_1d(FFTSIZE, (fftwf_complex *)vb->Comp, (fftwf_complex *)vb->Comp, FFTW_FORWARD, FFTW_MEASURE);
	PlanB = fftwf_plan_dft_1d(FFTSIZE, (fftwf_complex *)vb->Comp, (fftwf_complex *)vb->Comp, FFTW_BACKWARD, FFTW_MEASURE);
	if (atexit(freePlans))
		return INT_MIN;
#elif defined SFFT
	PlanF = sfft_init(FFTSIZE, SFFT_FORWARD);
	PlanB = sfft_init(FFTSIZE, SFFT_BACKWARD);
	if (!PlanF || !PlanB || atexit(freePlans))
		return INT_MIN;
#elif defined SPL
	iopt = 0;
	scale = 1.f;
	cfftc_(&iopt, &len, &scale, (complex *)vb->Comp, (complex *)vb->Comp, trigs, ifac, work, &lwork, &ierr);
	if (ierr)
		return ierr;
#elif defined VDSP
	fftData.realp = dataReal;
	fftData.imagp = dataImag;
	fftTemp.realp = tempReal;
	fftTemp.imagp = tempImag;
	if ((fftSetup = vDSP_create_fftsetup(FFTBITS, kFFTRadix2)) == NULL || atexit(freevDSP))
		return INT_MIN;
#elif defined ACML
	cfft1d(100, FFTSIZE, vb->Comp, comm, &info);
	if (info)
		return info;
#elif defined IPP
	{
		int nBufferSize;

		ippInit();
		if ((i = ippsFFTInitAlloc_C_32fc(&fftSpec, FFTBITS, IPP_FFT_DIV_FWD_BY_N, ippAlgHintFast)) < 0 || (i = ippsFFTGetBufSize_C_32fc(fftSpec, &nBufferSize)) < 0)
			return i;
		if ((fftBuf = malloc(nBufferSize)) == NULL || atexit(freeIPP))
			return INT_MIN;
	}
#endif
	return 0;
}
