/* METAXPON use example */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <time.h>
#include "metaxpon.h"

#ifndef min
#  define min(a,b)		(((a) < (b)) ? (a) : (b))
#endif

#define OUTBUF_SIZE_KB		16 /* [KB] see API description on its value */
#define INBUF_SIZE		(4 * 1024 * OUTBUF_SIZE_KB)
#define OUTBUF_SIZE		(1024 * OUTBUF_SIZE_KB)

typedef struct {		/* WAVE file header structure */
	DWORD main_chunk;	/* "RIFF" */
	DWORD length;		/* File length */
	DWORD chunk_type;	/* "WAVE" */
	DWORD sub_chunk;	/* "fmt " */
	DWORD sc_length;	/* Length of sub_chunk, can be 16, 18 or 20 */
	short format;		/* 1: signed integer, 3: IEEE 754 data */
	short modus;		/* 1: mono, 2: stereo */
	DWORD sample_fq;	/* Sampling frequency (e.g. 44100) */
	DWORD byte_p_sec;	/* Bytes per second (e.g. 176400) */
	short byte_p_spl;	/* Sample size: 1-24 bytes */
	short bit_p_spl;	/* Bits per sample: 8, 12, 16, 20, 24 or 32 */
	DWORD data_chunk;	/* "data" */
	DWORD data_length;	/* Data length */
} WavHdr;

/* WAVE file header constants for little- and big-endian machines */
#define RIFF_ID		0x46464952	/* "FFIR" */
#define WAVE_ID		0x45564157	/* "EVAW" */
#define FMT		0x20746D66	/* " tmf" */
#define DATA		0x61746164	/* "atad" */
#define STEREO		2

#define RIFF_ID_BIG	0x52494646	/* "RIFF" */
#define WAVE_ID_BIG	0x57415645	/* "WAVE" */
#define FMT_BIG		0x666D7420	/* "fmt " */
#define DATA_BIG	0x64617461	/* "data" */
#define STEREO_BIG	0x0200

#if FLOAT_DATA
#  ifdef __vax
#    include "convert_vax_data.h"	/* libvaxdata: http://pubs.usgs.gov/of/2005/1424 */
#  elif defined __VMS && !defined __IEEE_FLOAT
#    error Please specify the /FLOAT=IEEE_FLOAT compiler option
#  endif
#  define FORMAT_ID		3		    /* Normalised IEEE 754 data */
#  define SAMPLE_SIZE		(2 * sizeof(float)) /* 2 channels by 4 bytes */
#  define SAMPLE_BITS		32
#  define FORMAT_ID_BIG		0x0300
#  define SAMPLE_SIZE_BIG	0x0800
#  define SAMPLE_BITS_BIG	0x2000
#else
#  define FORMAT_ID		1		    /* Signed integer PCM data */
#  define SAMPLE_SIZE		(2 * sizeof(short)) /* 2 channels by 2 bytes */
#  define SAMPLE_BITS		16
#  define FORMAT_ID_BIG		0x0100
#  define SAMPLE_SIZE_BIG	0x0400
#  define SAMPLE_BITS_BIG	0x1000
#endif

#define BSwap32(x) (((x) >> 24) | (((x) & 0xFF0000) >> 8) | (((x) & 0xFF00) << 8) | ((x) << 24))

static void bswap(void *p, int n)
{
#if FLOAT_DATA
	DWORD *q = p;
	while (n--)		/* Reorder in reverse order n "float" numbers */
		q[n] = BSwap32(q[n]);
#else
	unsigned short *q = p;
	while (n--)		/* Swap even and odd bytes of n "short" words */
		q[n] = (unsigned short)((q[n] << 8) | (q[n] >> 8));
#endif
}

int main(int argc, char *argv[])
{
	long x = 1, bytesLeft, outBytes = 0;	/* Total bytes processed */
	int retval = -1, res = -1, bigEndian = *((char *)&x) == 0;
	size_t bytes;
#if FLOAT_DATA && defined __vax
	int n;
#endif
	double factor;
	BYTE *InBuffs, *OutBuff = NULL, *IntBuff = NULL;
	FILE *inf, *outf;
	clock_t start;
	static WavHdr header, hdrLEnd = {RIFF_ID, 0, WAVE_ID, FMT, 16, FORMAT_ID,
		STEREO, 44100, 4 * 44100, SAMPLE_SIZE, SAMPLE_BITS, 0, 0},
		hdrBigEnd = {RIFF_ID_BIG, 0, WAVE_ID_BIG, FMT_BIG, 0x10000000, FORMAT_ID_BIG,
		STEREO_BIG, 0x44AC0000, 0x10B10200, SAMPLE_SIZE_BIG, SAMPLE_BITS_BIG, 0, 0};
	WavHdr *hdr = bigEndian ? &hdrBigEnd : &hdrLEnd;
	static struct MXdata pd = {METAXPON_VERSION,0,0,0,0,{0},0,0,0,0,0,0,0,OUTBUF_SIZE_KB,0};

	/* Check the arguments */
	if (argc != 4) {
		puts("Invalid number of arguments\n"
		     "Usage:\t stretch <in_wave> <out_wave> <stretch_factor>\n"
		     "Example: stretch demo.wav slowdemo.wav 1.5 (slow down demo.wav 1.5 times)");
		return -4;
	}
	sscanf(argv[3], "%lf", &factor);
	if (factor < 0.5 || factor > 2.) {
		puts("Stretch factor out of range [0.5-2]");
		return -3;
	}
	/* Open the input file */
	if ((inf = fopen(argv[1], "rb")) == NULL) {
		printf("Can't open \"%s\"\n", argv[1]);
		return -2;
	}
	/* Check the WAVE header */
	if (fread(&header, sizeof(header), 1, inf) == 1) {
		hdr->length = header.length;	   /* Copy the real values */
		hdr->sc_length = header.sc_length; /* Could be 16, 18 or 20 */
		hdr->sample_fq = header.sample_fq;
		hdr->byte_p_sec = header.byte_p_sec;
		if (memcmp((const BYTE *)hdr, (const BYTE *)&header, offsetof(WavHdr, data_chunk)))
			goto bad;
		/* For sc_length > 16 find the first chunk and read its ID and size */
		if (fseek(inf, (bigEndian ? BSwap32(header.sc_length) : header.sc_length) + offsetof(WavHdr, format), SEEK_SET) || fread(&header.data_chunk, sizeof(DWORD), 2, inf) < 2)
			goto bad;
		/* Ignore chunks other than DATA */
		while (header.data_chunk != (DWORD)(bigEndian ? DATA_BIG : DATA))
			if (fseek(inf, bigEndian ? BSwap32(header.data_length) : header.data_length, SEEK_CUR) || fread(&header.data_chunk, sizeof(DWORD), 2, inf) < 2) /* Read data_chunk and data_length */
				goto bad;
	} else {
bad:		puts("Invalid or unsupported input WAVE file format");
		goto ret2;
	}
	/* Open the output file */
	if ((outf = fopen(argv[2], "wb")) == NULL) {
		printf("Can't open \"%s\"\n", argv[2]);
		goto ret2;
	}
	fwrite(&header, sizeof(header), 1, outf); /* Copy input to output header */

	/* Allocate memory for 2 input buffer halves, 1 output buffer */
	/* (enough for file operation), and the internal data buffers */
	if ((InBuffs = malloc(2 * INBUF_SIZE + 32)) == NULL || (OutBuff = malloc(OUTBUF_SIZE + 32)) == NULL || (IntBuff = malloc(INT_BUF_SIZE)) == NULL) {
		puts("Not enough memory for buffers");
		goto ret1;
	}
	/* Initialise METAXPON */
	pd.InBuf = InBuffs;		/* Set buffer addresses */
	pd.OutBuf = OutBuff;
	pd.Memory = IntBuff;
	if (MXinit(&pd) < 0) {		/* Always check the return value here */
		puts("Library version mismatch");
		goto ret1;
	}
	/* Prepare for start */
	pd.Tempo = (DWORD)(65536. / factor);/* Convert to 16.16; changeable anytime after MXinit */
	/* pd.Threshold = 0x18000;		"	"	"	"	"	"    */
	/* pd.MathUnit = 0;					"	"	"	"    */
	if (MXstart(&pd) < 0) {		/* Clear internal buffers and counters; set buffer sizes */
		puts("Incorrect output buffer size");
		goto ret1;
	}
	start = clock();

	/* Read both buffer halves at once. Assume that the input file is larger */
	/* than 2 * INBUF_SIZE, otherwise 'E0F' and 'Last' must be set (see API) */
	bytesLeft = (bigEndian ? BSwap32(header.data_length) : header.data_length) - fread(pd.InBuf, 1, 2 * INBUF_SIZE, inf);
#if FLOAT_DATA && defined __vax
	n = 2 * INBUF_SIZE / sizeof(float);
	to_vax_r4(pd.InBuf, pd.InBuf, &n);
#endif
	if (bigEndian)
		bswap(pd.InBuf, INBUF_SIZE / (1 + FLOAT_DATA));

	/* Time-scaling loop */
	while (res < 0) {		/* Last buffer processed */
		if (pd.Empty) {		/* Empty input buffer half */
			BYTE *p = (BYTE *)pd.InBuf + (1-pd.Half) * INBUF_SIZE;
			pd.Empty = 0;	/* Mark as filled */
			/* Empty input buffer half address = base + other half * buffer size */
			bytesLeft -= bytes = bytesLeft > 0 ? fread(p, 1, min(bytesLeft, INBUF_SIZE), inf) : 0; /* Don't read past chunk end (as if EOF) */
#if FLOAT_DATA && defined __vax
			n = bytes / sizeof(float);
			to_vax_r4(p, p, &n);
#endif
			if (bigEndian)
				bswap(p, bytes / (2 * (1 + FLOAT_DATA)));
			putchar('.');	/* Dots: the simplest progress indicator */
			fflush(stdout); /* For GCC */
			if (bytes < INBUF_SIZE) {	/* Partial buffer filled => end of file */
				/* End is in the just filled buffer half + 1; preserve other bits */
				pd.E0F = (BYTE)((pd.E0F & 0xFC) | (1 + 1-pd.Half));
				pd.Last = bytes;	/* Number of bytes in last buffer half */
			}
		}
		res = MXprocess(&pd); /* Do the time-scaling; append the data to the output file */
		bytes = (res == -1) ? OUTBUF_SIZE : SAMPLE_SIZE * res;
#if FLOAT_DATA && defined __vax
		n = bytes / sizeof(float);
		from_vax_r4(pd.OutBuf, pd.OutBuf, &n);
#endif
		if (bigEndian)
			bswap(pd.OutBuf, bytes / (2 * (1 + FLOAT_DATA)));
		outBytes += fwrite(pd.OutBuf, 1, bytes, outf); /* Bytes in the output buffer */
	}
	printf("\nElapsed time: %.3f s\n", (double)(clock() - start) / CLOCKS_PER_SEC);

	/* Adjust the output WAVE file header to the correct length */
	header.length = outBytes + sizeof(header) - 2 * sizeof(DWORD);
	if (bigEndian)
		header.length = BSwap32(header.length);
	header.data_length = bigEndian ? BSwap32(outBytes) : outBytes;
	header.sc_length = bigEndian ? 0x10000000 : 16; /* Could be 18 or 20 too */
	fseek(outf, 0, SEEK_SET);
	fwrite(&header, sizeof(header), 1, outf);
	retval = 0;
ret1:
	fclose(outf);
	free(InBuffs);
	free(OutBuff);
	free(IntBuff);
ret2:
	fclose(inf);
	return retval;
}
